import time
import asyncio
from copy import deepcopy
from types import MappingProxyType

from mona_sdk import MonaSingleMessage

from .exceptions import WrongOpenAIClassException
from .mona_client import get_mona_clients
from .util.func_util import (
    add_conditional_sampling,
    call_non_blocking_sync_or_async,
)
from .endpoints.completion import (
    COMPLETION_CLASS_NAME,
    get_completion_class,
    get_analysis_params as completion_analysis_getter,
    get_clean_message as completion_message_cleaner,
)
from .util.validation_util import validate_and_get_sampling_ratio

EMPTY_DICT = MappingProxyType({})

MONA_ARGS_PREFIX = "MONA_"
CONTEXT_ID_ARG_NAME = MONA_ARGS_PREFIX + "context_id"
EXPORT_TIMESTAMP_ARG_NAME = MONA_ARGS_PREFIX + "export_timestamp"
ADDITIONAL_DATA_ARG_NAME = MONA_ARGS_PREFIX + "additional_data"

# TODO(Itai): This is essetially a nice-looking "switch" statement. We should
#   try to use the name to find the exact monitoring-enrichment function and
#   filename instead of listing all options here.
ENDPOINT_NAME_TO_WRAPPER = {COMPLETION_CLASS_NAME: get_completion_class}


def _get_monitored_base_class(openai_class):
    """
    Returns a class that wrapps the given api class with that
    api-specific functionality.
    """
    class_name = openai_class.__name__

    if class_name not in ENDPOINT_NAME_TO_WRAPPER:
        raise WrongOpenAIClassException("Class not supported: " + class_name)
    return ENDPOINT_NAME_TO_WRAPPER[class_name](openai_class)


def _get_mona_single_message(
    api_name,
    request_input,
    start_time,
    is_exception,
    is_async,
    response,
    analysis_params_getter,
    specs,
    context_class,
    message_cleaner,
    additional_data,
    context_id,
    export_timestamp,
):
    """
    Returns a MonaSingleMessage object to be used for data
    exporting to Mona's servers by a Mona client.
    """

    message = {
        "input": request_input,
        "latency": time.time() - start_time,
        "is_exception": is_exception,
        "api_name": api_name,
        "is_async": is_async,
    }

    if additional_data:
        message["additional_data"] = additional_data

    if response:
        message["response"] = response
        message["analysis"] = analysis_params_getter(
            request_input, response, specs
        )

    message = message_cleaner(message, specs)

    return MonaSingleMessage(
        message=message,
        contextClass=context_class,
        contextId=context_id,
        exportTimestamp=export_timestamp,
    )


# TODO(itai): Consider creating some sturct (as NamedTuple or dataclass) for
#   the specs param.


def monitor(
    openai_class,
    mona_creds,
    context_class,
    specs=EMPTY_DICT,
    mona_clients_getter=get_mona_clients,
):
    """
    Returns a Wrapped version of a given OpenAI class with mona
    monitoring logic.
    This is the main exposed function of the mona_openai package and
    probably the only thing you need to use from this package.

    You can use the returned class' "create" and "acreate" functions
    exactly as you would the original class, and monitoring will be
    taken care of for you.

    This client will automatically monitor for you things like latency,
    prompt and response lengths, number of tokens, etc., along with any
    endpoint parameter usage (e.g., it tracks the "temperature" and
    "max_tokens" params you use).

    You can also add other named args when calling "create" or
    "acreate" by using a new named argument called
    "MONA_additional_data" and set it to any JSON serializable
    dictionary.
    This allows you to add metadata about the call such as a prompt
    template ID, information about the context in which the API call is
    made, etc...

    Furthermore, you can add to create/acreate functions mona specific
    arguments:
        MONA_context_id: The unique id of the context in which the call
            is made. By using this ID you can export more data to Mona
            to the same context from other places. If not used, the
            "id" field of the OpenAI Endpoint's response will be used.
        MONA_export_timestamp: Can be used to simulate as if the
            current call was made in a different time, as far as Mona
            is concerned.

    The returned monitored class also exposes a new class method called
    "get_mona_clients", which allows you to retrieve the clients and
    use them to export more data or communicate directly with Mona's
    API

    Read more about Mona and how to use it in Mona's docs on
    https://docs.monalabs.io.

    Args:
        openai_class: An OpenAI API class to wrap with monitoring
            capabilties.
        mona_creds: Either a dict or pair of Mona API key and secret to
            set up Mona's clients from its SDK
        context_class: The Mona context class name to use for
            monitoring. Use a name of your choice.
        specs: A dictionary of specifications such as monitoring
            sampling ratio.
        mona_clients_getter: Used only for testing purposes
    """
    client, async_client = mona_clients_getter(mona_creds)

    sampling_ratio = validate_and_get_sampling_ratio(specs)

    base_class = _get_monitored_base_class(openai_class)

    # TODO(itai): Add call to Mona servers to init the context class if it
    #   isn't inited yet once we have the relevant endpoint for this.

    class MonitoredOpenAI(base_class):
        """
        A mona-monitored version of an openai API class.
        """

        @classmethod
        def _get_mona_single_message(
            cls, kwargs_param, start_time, is_exception, is_async, response
        ):
            """
            Returns a MonaSingleMessage object to be used for data
            exporting to Mona's servers by a Mona client.
            """
            # Recreate the input dict to avoid manipulating the caller's data,
            # and remove Mona-related data.
            request_input = deepcopy(
                {
                    x: kwargs_param[x]
                    for x in kwargs_param
                    if not x.startswith(MONA_ARGS_PREFIX)
                }
            )

            return _get_mona_single_message(
                api_name=openai_class.__name__,
                request_input=request_input,
                start_time=start_time,
                is_exception=is_exception,
                is_async=is_async,
                response=response,
                analysis_params_getter=super()._get_analysis_params,
                specs=specs,
                context_class=context_class,
                message_cleaner=super()._get_clean_message,
                additional_data=kwargs_param.get(ADDITIONAL_DATA_ARG_NAME),
                context_id=kwargs_param.get(
                    CONTEXT_ID_ARG_NAME, response["id"] if response else None
                ),
                export_timestamp=kwargs_param.get(
                    EXPORT_TIMESTAMP_ARG_NAME, start_time
                ),
            )

        @classmethod
        async def _inner_create(
            cls, export_function, super_function, args, kwargs
        ):
            """
            The main logic for wrapping create functions with mona data
            exporting.
            This internal function porovides a template for both sync
            and async activations (helps with wrapping both "create"
            and "acreate").
            """
            start_time = time.time()

            response = None

            async def _inner_mona_export(is_exception):
                return await call_non_blocking_sync_or_async(
                    export_function,
                    (
                        cls._get_mona_single_message(
                            kwargs,
                            start_time,
                            is_exception,
                            super_function.__name__ == "acreate",
                            response,
                        ),
                    ),
                )

            mona_export = add_conditional_sampling(
                _inner_mona_export, sampling_ratio
            )

            try:
                # Call the actual openai create function without the Mona
                # specific arguments.
                response = await call_non_blocking_sync_or_async(
                    super_function,
                    args,
                    {
                        x: kwargs[x]
                        for x in kwargs
                        if not x.startswith(MONA_ARGS_PREFIX)
                    },
                )
            except Exception:
                if not specs.get("avoid_monitoring_exceptions", False):
                    await mona_export(True)
                raise

            await mona_export(False)

            return response

        @classmethod
        def create(cls, *args, **kwargs):
            """
            A mona-monitored version of the openai base class' "create"
            function.
            """
            return asyncio.run(
                cls._inner_create(client.export, super().create, args, kwargs)
            )

        @classmethod
        async def acreate(cls, *args, **kwargs):
            """
            An async mona-monitored version of the openai base class'
            "acreate" function.
            """
            return await cls._inner_create(
                async_client.export_async, super().acreate, args, kwargs
            )

        @classmethod
        def get_mona_clients(cls):
            """
            Returns the two Mona clients this class works with to allow
            exporting more data or communicating directly with Mona's
            API.
            """
            return (client, async_client)

    return type(base_class.__name__, (MonitoredOpenAI,), {})


def get_rest_monitor(
    openai_endpoint_name,
    mona_creds,
    context_class,
    specs=EMPTY_DICT,
    mona_clients_getter=get_mona_clients,
):
    """
    Returns a client class for monitoring OpenAI REST calls not done
    using the OpenAI python client (e.g., for Azure users using their
    endpoints directly). This isn't a wrapper for any http requesting
    library and doesn't call the OpenAI API for you - it's just an easy
    logging client to log requests, responses and exceptions.
    """

    # TODO(itai): Consider creating an async version as well.
    client, _ = mona_clients_getter(mona_creds)

    sampling_ratio = validate_and_get_sampling_ratio(specs)

    class RestClient:
        """
        This will be the returned Mona logging class. We follow
        OpenAI's way of doing things by using a static classe with
        relevant class methods.
        """

        @classmethod
        def log_request(
            cls,
            request_dict,
            additional_data=None,
            context_id=None,
            export_timestamp=None,
        ):
            """
            This function should be called with a request data dict,
            for example, what you would use as "json" when using
            "requests" to post.

            It returns a response logging function to be used with the
            response object.
            """
            start_time = time.time()

            inner_response = None

            def _inner_mona_export(is_exception):
                return client.export(
                    _get_mona_single_message(
                        api_name=openai_endpoint_name,
                        request_input=request_dict,
                        start_time=start_time,
                        is_exception=is_exception,
                        is_async=False,
                        response=inner_response,
                        analysis_params_getter=completion_analysis_getter,
                        specs=specs,
                        context_class=context_class,
                        message_cleaner=completion_message_cleaner,
                        additional_data=additional_data,
                        context_id=context_id,
                        export_timestamp=export_timestamp,
                    )
                )

            mona_export = add_conditional_sampling(
                _inner_mona_export, sampling_ratio
            )

            def log_response(response):
                """
                Only when this function is called, will data be logged
                out to Mona. This function should be called with a
                response object from the OpenAI API as close as
                possible to when it is received to allow accurate
                latency logging.
                """
                nonlocal inner_response
                inner_response = response
                return mona_export(False)

            def log_exception():
                return mona_export(True)

            return log_response, log_exception

        @classmethod
        def get_mona_client(cls):
            """
            Returns the two Mona client this class works with to allow
            exporting more data or communicating directly with Mona's
            API.
            """
            return client

    return RestClient



"""
A module with utility (usually higher order) functions for enriching
and extending other functions' functionalities.
"""
from types import MappingProxyType
from random import random
from inspect import iscoroutinefunction

EMPTY_DICT = MappingProxyType({})


async def call_non_blocking_sync_or_async(
    function, func_args=(), func_kwargs=EMPTY_DICT
):
    """
    A higher order function that allows calling both sync and async
    functions as if they were async, avoid blocking when relevant, and
    maintain one code base for both cases.
    """
    if iscoroutinefunction(function):
        return await function(*func_args, **func_kwargs)
    return function(*func_args, **func_kwargs)


def add_conditional_sampling(inner_func, sampling_ratio):
    """
    A higher order function that returns a "sampled" version of the
    given inner function only if needed. This allows for adding
    sampling mechanisms while avoiding conditionals or random number
    creations when either is not necessary.
    """

    async def _sampled_func(*args, **kwargs):
        if random() < sampling_ratio:
            return await inner_func(*args, **kwargs)

    return inner_func if sampling_ratio == 1 else _sampled_func



"""
Logic to create profanity analysis.
"""

from profanity_check import predict_prob


def _get_profanity_prob(texts):
    return tuple(round(x, 2) for x in predict_prob(texts))


def get_full_profainty_analysis(prompt, answers):
    return {
        "prompt_profanity_prob": _get_profanity_prob((prompt,))[0],
        "answer_profanity_prob": _get_profanity_prob(answers),
    }



"""
A module to derive text-related metrics such as text length, usage of
specific grammatical words, text repetition, etc...

These analyses can be used to detect significant drifts that could be
caused by hallucinations or bugs.

NOTE: There are many more analyses that can be added here.
"""

from ..util.math import get_quotients

PREPOSITIONS = set(
    (
        "aboard",
        "about",
        "above",
        "across",
        "after",
        "against",
        "along",
        "amid",
        "among",
        "around",
        "as",
        "at",
        "before",
        "behind",
        "below",
        "beneath",
        "beside",
        "between",
        "beyond",
        "but",
        "by",
        "concerning",
        "considering",
        "despite",
        "down",
        "during",
        "except",
        "for",
        "from",
        "in",
        "inside",
        "into",
        "like",
        "near",
        "of",
        "off",
        "on",
        "onto",
        "out",
        "outside",
        "over",
        "past",
        "regarding",
        "round",
        "since",
        "through",
        "throughout",
        "till",
        "to",
        "toward",
        "under",
        "underneath",
        "until",
        "unto",
        "up",
        "upon",
        "with",
        "within",
        "without",
    )
)


def _count_prepositions(splitted_text):
    """
    Returns the number of function words in a given text.
    """
    return len([x for x in splitted_text if x in PREPOSITIONS])


def _count_words_not_in_prompt_for_single_answer(
    prompt_words, splitted_answer
):
    return len([word for word in splitted_answer if word not in prompt_words])


def _count_words_not_in_prompt(splitted_prompt, splitted_answers):
    """
    for each answer, returns the number of words it has that do not
    exist in the prompt. High number of such "new" words can be a
    symptom of hallucinations.
    """
    prompt_words = set(splitted_prompt)
    return tuple(
        _count_words_not_in_prompt_for_single_answer(
            prompt_words, splitted_answer
        )
        for splitted_answer in splitted_answers
    )


def get_full_textual_analysis(prompt, answers):
    """
    Provides a dictionary with full textual analyses metrics about the
    prompt, the answers and their relationship.
    """
    splitted_prompt = prompt.split()
    splitted_answers = tuple(answer.split() for answer in answers)

    prompt_preposition_count = _count_prepositions(splitted_prompt)
    answer_preposition_counts = tuple(
        _count_prepositions(splitted_answer)
        for splitted_answer in splitted_answers
    )
    answer_words_counts = tuple(
        len(splitted_answer) for splitted_answer in splitted_answers
    )
    answer_words_not_in_prompt_count = _count_words_not_in_prompt(
        splitted_prompt, splitted_answers
    )

    return {
        "prompt_length": len(prompt),
        "answer_length": tuple(len(answer) for answer in answers),
        "prompt_word_count": len(splitted_prompt),
        "answer_word_count": answer_words_counts,
        "prompt_preposition_count": prompt_preposition_count,
        "prompt_preposition_ratio": prompt_preposition_count
        / len(splitted_prompt)
        if prompt_preposition_count != 0
        else 0,
        "answer_preposition_count": answer_preposition_counts,
        "answer_preposition_ratio": get_quotients(
            answer_preposition_counts, answer_words_counts
        ),
        "answer_words_not_in_prompt_count": answer_words_not_in_prompt_count,
        "answer_words_not_in_prompt_ratio": get_quotients(
            answer_words_not_in_prompt_count, answer_words_counts
        ),
    }



"""
A module to gather analysis metrics from given prompts and and answers.

NOTE: It remains to be seen if the functionality here should be
    further refactored to better fit more endpoints besides
    "Completion".

TODO(itai): Add more analysis categories.
"""

from .privacy import get_full_privacy_analysis
from .textual import get_full_textual_analysis
from .profanity import get_full_profainty_analysis

ANALYSIS_TYPES_TO_FUNCTION = {
    "privacy": get_full_privacy_analysis,
    "textual": get_full_textual_analysis,
    "profanity": get_full_profainty_analysis,
}


def get_full_analysis(prompt, answers, specs):
    """
    Returns a dict mapping each analysis type to all related analysis
    fields for the given prompt and answers according to the given
    specs (if no "analysis" spec is given - return result for all
    analysis types).

    TODO(itai): Consider propogating the specs to allow the user to
        choose specific anlyses to be made from within each analysis
        category.
    """
    return {
        x: ANALYSIS_TYPES_TO_FUNCTION[x](prompt, answers)
        for x in ANALYSIS_TYPES_TO_FUNCTION
        if specs.get("analysis", {}).get(x, True)
    }



from mona_sdk.client import Client
from mona_sdk.async_client import AsyncClient

from .exceptions import InvalidMonaCredsException


MONA_API_KEY_KEY = "key"
MONA_API_SECRET_KEY = "secret"


def get_mona_clients(creds):
    """
    Returns both a sync and an async mona client for the given
    credentials.

    creds: Either a tuple or a dict containing API key and secret for
        Mona's API.
    """
    if len(creds) != 2:
        raise InvalidMonaCredsException(
            "There should be exactly two parts to Mona creds. API key and"
            " secret."
        )

    # Creds could be in a dict.
    if isinstance(creds, dict):
        if MONA_API_KEY_KEY not in creds or MONA_API_SECRET_KEY not in creds:
            raise InvalidMonaCredsException(
                f"Mona creds dict should hold keys:"
                f"{MONA_API_KEY_KEY}, {MONA_API_SECRET_KEY}"
            )
        return Client(
            creds[MONA_API_KEY_KEY], creds[MONA_API_SECRET_KEY]
        ), AsyncClient(creds[MONA_API_KEY_KEY], creds[MONA_API_SECRET_KEY])

    # If creds are not in a dict, they are in a tuple (pair).
    if not isinstance(creds[0], str) or not isinstance(creds[1], str):
        raise InvalidMonaCredsException(
            "Mona API key and secret should both be strings."
        )

    return Client(creds[0], creds[1]), AsyncClient(creds[0], creds[1])



"""
Utility module for math-related functionality.
"""


class InvalidQuotientssInputError(Exception):
    pass


def get_quotients(iter_1, iter_2):
    """
    For two given iterables of numbers, returns a tuple of the factors
    between each pair of numbers from the two lists.
    """
    iter_1, iter_2 = tuple(iter_1), tuple(iter_2)
    if len(iter_1) != len(iter_2):
        raise InvalidQuotientssInputError()
    return tuple(x / y if y != 0 else None for x, y in zip(iter_1, iter_2))



"""
The Mona wrapping code for OpenAI's Completion API.
"""
from copy import deepcopy

from ..analysis.analysis import get_full_analysis
from ..util.validation_util import validate_openai_class

COMPLETION_CLASS_NAME = "Completion"


def get_analysis_params(input, response, specs):
    """
    Returns the full dict of analysis parameters for the given input
    and response.
    """
    return get_full_analysis(
        prompt=input["prompt"],
        answers=tuple(choice["text"] for choice in response["choices"]),
        specs=specs,
    )


def get_clean_message(message, specs):
    """
    Returns a copy of the given message with relevant data removed, for
    example the actual texts, to avoid sending such information, that
    is sometimes sensitive, to Mona.
    """
    new_message = deepcopy(message)
    if not specs.get("export_prompt", False):
        new_message["input"].pop("prompt")

    if "response" in message and not specs.get("export_response_texts", False):
        for choice in new_message["response"]["choices"]:
            choice.pop("text")

    return new_message


def get_completion_class(openai_class):
    """
    Returns a monitored class wrapping the given openai class, which is
    expected to be a "openai.Completion" class, enriching it with
    specific capabilities to be used by an inhereting monitored class.
    """
    validate_openai_class(openai_class, COMPLETION_CLASS_NAME)

    class MonitoredCompletion(openai_class):
        @classmethod
        def _get_analysis_params(cls, input, response, specs):
            return get_analysis_params(input, response, specs)

        @classmethod
        def _get_clean_message(cls, message, specs):
            return get_clean_message(message, specs)

    return MonitoredCompletion



"""
Functionality for extracting privacy information from GAI responses
when comparing to given input prompts.

TODO(itai): Add many more functions to extract information such as:
    Bank accounts, full names, SSN, passport numbers, etc...
"""
import re

from phonenumbers import PhoneNumberMatcher

# TODO(itai): Add module-level tests for this module, specifically for email
#   extraction, since this is our own logic and not using an external library.

EMAIL_RE_PATTERN = (
    r"([-!#-'*+/-9=?A-Z^-~]+(\.[-!#-'*+/-9=?A-Z^-~]+)*|\"([]!#-[^-~"
    r" \t]|(\\[\t -~]))+\")@([-!#-'*+/-9=?A-Z^-~]+(\.[-!#-'*+/-9=?A"
    r"-Z^-~]+)*|\[[\t -Z^-~]*])"
)


def _count_unknowns(prompt, answer, extract_func):
    return len(extract_func(prompt) - extract_func(answer))


def _extract_phone_numbers(prompt):
    """
    Extract phone numbers from a prompt string and return as a set.
    """
    phone_numbers = set()
    # We use "US" just as a default region in case there are no country codes
    # since we don't care about the formatting of the found number, but just
    # whether it is a phone number or not, this has no consequences.
    for match in PhoneNumberMatcher(prompt, "US"):
        number_string = "+{}{}".format(
            match.number.country_code, match.number.national_number
        )
        phone_numbers.add(number_string)
    return phone_numbers


def _count_unknown_phone_numbers(prompt, answer):
    """
    Count the number of phone numbers that appear in the prompt but
    not in the input prompt.
    """
    return _count_unknowns(prompt, answer, _extract_phone_numbers)


def _extract_all_emails(prompt):
    """
    returns all email addresses found in the given prompt.
    """
    return set(re.findall(EMAIL_RE_PATTERN, prompt))


def _count_unknown_emails(prompt, answer):
    """
    Returns the number of email addresses found in prompt and not in
    the answer param.
    """
    return _count_unknowns(prompt, answer, _extract_all_emails)


def get_full_privacy_analysis(prompt, answers):
    """
    Returns a dictionary with a full privacy analysis of the given
    prompt, when comparing to the "answer" param.
    """
    # TODO(itai): Extract the phone numbers from the prompt just once
    #   to improve efficiency.
    return {
        "prompt_phone_number_count": len(_extract_phone_numbers(prompt)),
        "answer_unknown_phone_number_count": tuple(
            _count_unknown_phone_numbers(prompt, answer) for answer in answers
        ),
        "prompt_email_count": len(_extract_all_emails(prompt)),
        "answer_unkown_email_count": tuple(
            _count_unknown_emails(prompt, answer) for answer in answers
        ),
    }
